mod alternate_parser;
mod left_right_parsers;
mod many_parsers;
mod separated_parsers;
mod tuple_parser;

use crate::combinators::alternate_parser::AlternateParser;
use crate::combinators::left_right_parsers::{LeftParser, RightParser};
use crate::combinators::many_parsers::{
    Many1Parser, Many1TillParser, ManyParser, ManyTillParser, Skip1Parser, SkipParser,
};
use crate::combinators::separated_parsers::{SeparateBy1Parser, SeparateOrEndBy1Parser};
use crate::combinators::tuple_parser::ParserTuple;
use crate::parser::{any, Parser, ParserContext, ParserResult};
use std::cell::RefCell;
use std::fmt::Debug;
use std::marker::PhantomData;
use std::rc::Rc;

pub trait ParserExt<TToken, T, TContext>: Parser<TToken, T, TContext>
where
    TToken: Debug + Clone,
{
    fn left<TRight, PRight>(self, right_parser: PRight) -> LeftParser<Self, PRight, TRight>
    where
        Self: Sized,
        PRight: Parser<TToken, TRight, TContext>,
    {
        LeftParser {
            left_parser: self,
            right_parser,
            phantom_data: PhantomData,
        }
    }

    fn right<TRight, PRight>(self, right_parser: PRight) -> RightParser<Self, PRight, T>
    where
        Self: Sized,
        PRight: Parser<TToken, TRight, TContext>,
    {
        RightParser {
            left_parser: self,
            right_parser,
            phantom_data: PhantomData,
        }
    }

    fn alternate<P>(self, second_parser: P) -> AlternateParser<Self, P>
    where
        Self: Sized,
        P: Parser<TToken, T, TContext>,
    {
        AlternateParser {
            first: self,
            second: second_parser,
        }
    }

    fn many(self) -> ManyParser<Self>
    where
        Self: Sized,
    {
        ManyParser { parser: self }
    }

    fn many1(self) -> Many1Parser<Self>
    where
        Self: Sized,
    {
        Many1Parser { parser: self }
    }

    fn skip(self) -> SkipParser<Self, T>
    where
        Self: Sized,
    {
        SkipParser {
            parser: self,
            phantom_data: PhantomData,
        }
    }

    fn skip1(self) -> Skip1Parser<Self, T>
    where
        Self: Sized,
    {
        Skip1Parser {
            parser: self,
            phantom_data: PhantomData,
        }
    }

    fn many_till<TTerminator, PTerminator>(
        self,
        terminator: PTerminator,
    ) -> ManyTillParser<Self, PTerminator, TTerminator>
    where
        Self: Sized,
        PTerminator: Parser<TToken, TTerminator, TContext>,
    {
        ManyTillParser {
            parser: self,
            terminator,
            phantom_data: PhantomData,
        }
    }

    fn many1_till<TTerminator, PTerminator>(
        self,
        terminator: PTerminator,
    ) -> Many1TillParser<Self, PTerminator, TTerminator>
    where
        Self: Sized,
        PTerminator: Parser<TToken, TTerminator, TContext>,
    {
        Many1TillParser {
            parser: self,
            terminator,
            phantom_data: PhantomData,
        }
    }
}

impl<TToken, T, TContext, P> ParserExt<TToken, T, TContext> for P
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
{
}

fn alternate_helper<TToken, T, TContext, P1, P2>(first: P1, second: P2) -> AlternateParser<P1, P2>
where
    TToken: Debug + Clone,
    P1: Parser<TToken, T, TContext>,
    P2: Parser<TToken, T, TContext>,
{
    AlternateParser { first, second }
}

#[macro_export]
macro_rules! alternate {
    ($first:expr, $( $second:expr ),*) => {
        {
            let parser = $first;
            $(
                let parser = crate::combinators::alternate_helper(parser, $second);
            )*
            parser
        }
    };
}

pub fn tuple<TToken, T, TContext, TList>(
    list: TList,
) -> impl Fn(Rc<RefCell<ParserContext<TToken, TContext>>>, &[TToken]) -> ParserResult<TToken, T>
where
    TToken: Debug + Clone,
    TList: ParserTuple<TToken, T, TContext>,
{
    move |context, input| list.parse(context, input)
}

pub fn take_till<TToken, TTerminator, TContext, PTerminator>(
    termiantor: PTerminator,
) -> impl Parser<TToken, Vec<TToken>, TContext>
where
    TToken: Debug + Clone,
    PTerminator: Parser<TToken, TTerminator, TContext>,
{
    any().many_till(termiantor)
}

pub fn take1_till<TToken, TTerminator, TContext, PTerminator>(
    terminator: PTerminator,
) -> impl Parser<TToken, Vec<TToken>, TContext>
where
    TToken: Debug + Clone,
    PTerminator: Parser<TToken, TTerminator, TContext>,
{
    any().many1_till(terminator)
}

pub fn quote<TToken, T, TLeft, TRight, TContext, PLeft, P, PRight>(
    left: PLeft,
    parser: P,
    right: PRight,
) -> impl Parser<TToken, Vec<T>, TContext>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
    PLeft: Parser<TToken, TLeft, TContext>,
    PRight: Parser<TToken, TRight, TContext>,
{
    left.right(parser.many_till(right))
}

pub fn separate_by1<TToken, T, TSeparator, TContext, P, PSeparator>(
    parser: P,
    separator: PSeparator,
) -> impl Parser<TToken, Vec<T>, TContext>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
    PSeparator: Parser<TToken, TSeparator, TContext>,
{
    SeparateBy1Parser {
        parser,
        separator,
        phantom_data: PhantomData,
    }
}

pub fn separate_by<TToken, T, TSeparator, TContext, P, PSeparator>(
    parser: P,
    separator: PSeparator,
) -> impl Parser<TToken, Vec<T>, TContext>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
    PSeparator: Parser<TToken, TSeparator, TContext>,
{
    separate_by1(parser, separator).next(|r| match r {
        Err((input, _)) => Ok((input, vec![])),
        r => r,
    })
}

pub fn end_by1<TToken, T, TSeparator, TContext, P, PSeparator>(
    parser: P,
    separator: PSeparator,
) -> impl Parser<TToken, Vec<T>, TContext>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
    PSeparator: Parser<TToken, TSeparator, TContext>,
{
    parser.many1_till(separator)
}

pub fn end_by<TToken, T, TSeparator, TContext, P, PSeparator>(
    parser: P,
    separator: PSeparator,
) -> impl Parser<TToken, Vec<T>, TContext>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
    PSeparator: Parser<TToken, TSeparator, TContext>,
{
    parser.many_till(separator)
}

pub fn separate_or_end_by1<TToken, T, TSeparator, TContext, P, PSeparator>(
    parser: P,
    separator: PSeparator,
) -> impl Parser<TToken, Vec<T>, TContext>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
    PSeparator: Parser<TToken, TSeparator, TContext>,
{
    SeparateOrEndBy1Parser {
        parser,
        separator,
        phantom_data: PhantomData,
    }
}

pub fn separate_or_end_by<TToken, T, TSeparator, TContext, P, PSeparator>(
    parser: P,
    separator: PSeparator,
) -> impl Parser<TToken, Vec<T>, TContext>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
    PSeparator: Parser<TToken, TSeparator, TContext>,
{
    separate_or_end_by1(parser, separator).next(|r| match r {
        Err((i, _)) => Ok((i, vec![])),
        r => r,
    })
}
