use crate::parser::{Parser, ParserContext, ParserResult};
use std::cell::RefCell;
use std::fmt::Debug;
use std::rc::Rc;

pub struct AlternateParser<P1, P2> {
    pub(crate) first: P1,
    pub(crate) second: P2,
}

impl<TToken, T, TContext, P1, P2> Parser<TToken, T, TContext> for AlternateParser<P1, P2>
where
    TToken: Debug + Clone,
    P1: Parser<TToken, T, TContext>,
    P2: Parser<TToken, T, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T> {
        match self.first.parse(context.clone(), input) {
            Err((input, _)) => self.second.parse(context, input),
            r => r,
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::alternate;
    use crate::combinators::ParserExt;
    use crate::parser::parser_test_helper;
    use crate::text::char_parser;

    #[test]
    fn alternate_test() {
        fn parser(
            context: Rc<RefCell<ParserContext<char, ()>>>,
            input: &[char],
        ) -> ParserResult<char, char> {
            char_parser('a')
                .alternate(char_parser('b'))
                .alternate(char_parser('c'))
                .parse(context, input)
        }

        parser_test_helper(ParserContext::new_with_str("a", ()), &'a', parser);
        parser_test_helper(ParserContext::new_with_str("b", ()), &'b', parser);
        parser_test_helper(ParserContext::new_with_str("c", ()), &'c', parser);
    }

    #[test]
    fn alternate_macro_test() {
        fn parser(
            context: Rc<RefCell<ParserContext<char, ()>>>,
            input: &[char],
        ) -> ParserResult<char, char> {
            alternate!(char_parser('a'), char_parser('b'), char_parser('c')).parse(context, input)
        }

        parser_test_helper(ParserContext::new_with_str("a", ()), &'a', parser);
        parser_test_helper(ParserContext::new_with_str("b", ()), &'b', parser);
        parser_test_helper(ParserContext::new_with_str("c", ()), &'c', parser);
    }
}
