use crate::combinators::ParserExt;
use crate::parser::{Parser, ParserContext, ParserResult};
use std::cell::RefCell;
use std::fmt::Debug;
use std::marker::PhantomData;
use std::rc::Rc;

pub struct LeftParser<PLeft, PRight, TRight> {
    pub(crate) left_parser: PLeft,
    pub(crate) right_parser: PRight,
    pub(crate) phantom_data: PhantomData<TRight>,
}

impl<TToken, T, TContext, TRight, PLeft, PRight> Parser<TToken, T, TContext>
    for LeftParser<PLeft, PRight, TRight>
where
    TToken: Debug + Clone,
    PLeft: Parser<TToken, T, TContext>,
    PRight: Parser<TToken, TRight, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T> {
        let (input, r) = self.left_parser.parse(context.clone(), input)?;

        match self.right_parser.parse(context, input) {
            Ok((input, _)) => Ok((input, r)),
            Err(r) => Err(r),
        }
    }
}

pub struct RightParser<PLeft, PRight, T> {
    pub(crate) left_parser: PLeft,
    pub(crate) right_parser: PRight,
    pub(crate) phantom_data: PhantomData<T>,
}

impl<TToken, T, TContext, TRight, PLeft, PRight> Parser<TToken, TRight, TContext>
    for RightParser<PLeft, PRight, T>
where
    TToken: Debug + Clone,
    PLeft: Parser<TToken, T, TContext>,
    PRight: Parser<TToken, TRight, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, TRight> {
        let (input, _) = self.left_parser.parse(context.clone(), input)?;

        self.right_parser.parse(context, input)
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::parser::{parser_test_helper, ParserContext};
    use crate::text::char_parser;

    #[test]
    fn left_test() {
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &'a',
            char_parser('a').left(char_parser('b')),
        );
    }

    #[test]
    fn right_test() {
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &'b',
            char_parser('a').right(char_parser('b')),
        )
    }
}
