use crate::parser::{Parser, ParserContext, ParserResult};
use std::cell::RefCell;
use std::fmt::Debug;
use std::marker::PhantomData;
use std::rc::Rc;

pub struct SeparateBy1Parser<P, PSeparator, TSeparator> {
    pub(crate) parser: P,
    pub(crate) separator: PSeparator,
    pub(crate) phantom_data: PhantomData<TSeparator>,
}

impl<TToken, T, TSeparator, TContext, P, PSeparator> Parser<TToken, Vec<T>, TContext>
    for SeparateBy1Parser<P, PSeparator, TSeparator>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
    PSeparator: Parser<TToken, TSeparator, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, Vec<T>> {
        let (mut input, first) = self.parser.parse(context.clone(), input)?;
        let mut result = vec![first];

        loop {
            if let Ok((i, _)) = self.separator.parse(context.clone(), input) {
                let (i, r) = self.parser.parse(context.clone(), i)?;
                input = i;
                result.push(r);
            } else {
                break;
            }
        }

        Ok((input, result))
    }
}

pub struct SeparateOrEndBy1Parser<P, PSeparator, TSeparator> {
    pub(crate) parser: P,
    pub(crate) separator: PSeparator,
    pub(crate) phantom_data: PhantomData<TSeparator>,
}

impl<TToken, T, TSeparator, TContext, P, PSeparator> Parser<TToken, Vec<T>, TContext>
    for SeparateOrEndBy1Parser<P, PSeparator, TSeparator>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
    PSeparator: Parser<TToken, TSeparator, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, Vec<T>> {
        let (mut input, first) = self.parser.parse(context.clone(), input)?;
        let mut result = vec![first];

        loop {
            if let Ok((i, _)) = self.separator.parse(context.clone(), input) {
                if let Ok((i, r)) = self.parser.parse(context.clone(), i) {
                    result.push(r);
                    input = i;
                } else {
                    break;
                }
            } else {
                break;
            }
        }

        Ok((input, result))
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::combinators::{
        end_by, end_by1, quote, separate_by, separate_by1, separate_or_end_by, separate_or_end_by1,
    };
    use crate::parser::{failed_parser_test_helper, parser_test_helper, satisfy};
    use crate::text::char_parser;

    #[test]
    fn quote_test() {
        fn parser(
            context: Rc<RefCell<ParserContext<char, ()>>>,
            input: &[char],
        ) -> ParserResult<char, Vec<char>> {
            quote(
                char_parser('\''),
                satisfy(|c: &char| c.is_ascii()),
                char_parser('\''),
            )
            .parse(context, input)
        }

        parser_test_helper(
            ParserContext::new_with_str("'abc'", ()),
            &vec!['a', 'b', 'c'],
            parser,
        );
        parser_test_helper(ParserContext::new_with_str("''", ()), &vec![], parser);
        failed_parser_test_helper(ParserContext::new_with_str("asd", ()), parser)
    }

    #[test]
    fn separate_by1_test() {
        parser_test_helper(
            ParserContext::new_with_str("a,b,c", ()),
            &vec!['a', 'b', 'c'],
            separate_by1(satisfy(|c: &char| c.is_ascii()), char_parser(',')),
        );
        parser_test_helper(
            ParserContext::new_with_str("a", ()),
            &vec!['a'],
            separate_by1(satisfy(|c: &char| c.is_ascii()), char_parser(',')),
        );
        parser_test_helper(
            ParserContext::new_with_str("bbc", ()),
            &vec!['b'],
            separate_by1(satisfy(|c: &char| c.is_ascii()), char_parser(',')),
        );
    }

    #[test]
    fn separate_by_test() {
        parser_test_helper(
            ParserContext::new_with_str("1,2,3", ()),
            &vec!['1', '2', '3'],
            separate_by(satisfy(|c: &char| c.is_ascii()), char_parser(',')),
        );
        parser_test_helper(
            ParserContext::new_with_str("1", ()),
            &vec!['1'],
            separate_by(satisfy(|c: &char| c.is_ascii()), char_parser(',')),
        );
        parser_test_helper(
            ParserContext::new_with_str("", ()),
            &vec![],
            separate_by(satisfy(|c: &char| c.is_ascii()), char_parser(',')),
        );
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &vec!['a'],
            separate_by(satisfy(|c: &char| c.is_ascii()), char_parser(',')),
        );
    }

    #[test]
    fn end_by1_test() {
        parser_test_helper(
            ParserContext::new_with_str("aab", ()),
            &vec!['a', 'a'],
            end_by1(char_parser('a'), char_parser('b')),
        );
        parser_test_helper(
            ParserContext::new_with_str("ab", ()),
            &vec!['a'],
            end_by1(char_parser('a'), char_parser('b')),
        );
        failed_parser_test_helper(
            ParserContext::new_with_str("b", ()),
            end_by1(char_parser('a'), char_parser('b')),
        );
        failed_parser_test_helper(
            ParserContext::new_with_str("cd", ()),
            end_by1(char_parser('a'), char_parser('b')),
        );
    }

    #[test]
    fn end_by_test() {
        parser_test_helper(
            ParserContext::new_with_str("aab", ()),
            &vec!['a', 'a'],
            end_by(char_parser('a'), char_parser('b')),
        );
        parser_test_helper(
            ParserContext::new_with_str("ab", ()),
            &vec!['a'],
            end_by(char_parser('a'), char_parser('b')),
        );
        parser_test_helper(
            ParserContext::new_with_str("b", ()),
            &vec![],
            end_by(char_parser('a'), char_parser('b')),
        );
        failed_parser_test_helper(
            ParserContext::new_with_str("cd", ()),
            end_by(char_parser('a'), char_parser('b')),
        );
    }

    #[test]
    fn separate_or_end_by1_test() {
        parser_test_helper(
            ParserContext::new_with_str("1,2,3,", ()),
            &vec!['1', '2', '3'],
            separate_or_end_by1(satisfy(|c: &char| c.is_numeric()), char_parser(',')),
        );
        parser_test_helper(
            ParserContext::new_with_str("1,", ()),
            &vec!['1'],
            separate_or_end_by1(satisfy(|c: &char| c.is_numeric()), char_parser(',')),
        );
        parser_test_helper(
            ParserContext::new_with_str("1,2,3", ()),
            &vec!['1', '2', '3'],
            separate_or_end_by1(satisfy(|c: &char| c.is_numeric()), char_parser(',')),
        );
        parser_test_helper(
            ParserContext::new_with_str("1", ()),
            &vec!['1'],
            separate_or_end_by1(satisfy(|c: &char| c.is_numeric()), char_parser(',')),
        );
        failed_parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            separate_or_end_by1(satisfy(|c: &char| c.is_numeric()), char_parser(',')),
        );
    }

    #[test]
    fn separate_or_end_by_test() {
        parser_test_helper(
            ParserContext::new_with_str("1,2,3,", ()),
            &vec!['1', '2', '3'],
            separate_or_end_by(satisfy(|c: &char| c.is_numeric()), char_parser(',')),
        );
        parser_test_helper(
            ParserContext::new_with_str("1,", ()),
            &vec!['1'],
            separate_or_end_by(satisfy(|c: &char| c.is_numeric()), char_parser(',')),
        );
        parser_test_helper(
            ParserContext::new_with_str("1,2,3", ()),
            &vec!['1', '2', '3'],
            separate_or_end_by(satisfy(|c: &char| c.is_numeric()), char_parser(',')),
        );
        parser_test_helper(
            ParserContext::new_with_str("1", ()),
            &vec!['1'],
            separate_or_end_by(satisfy(|c: &char| c.is_numeric()), char_parser(',')),
        );
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &vec![],
            separate_or_end_by(satisfy(|c: &char| c.is_numeric()), char_parser(',')),
        );
    }
}
