use crate::parser::{Parser, ParserContext, ParserResult};
use std::cell::RefCell;
use std::fmt::Debug;
use std::rc::Rc;

pub trait ParserTuple<TToken, T, TContext>
where
    TToken: Debug + Clone,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T>;
}

impl<TToken, T1, T2, TContext, P1, P2> ParserTuple<TToken, (T1, T2), TContext> for (P1, P2)
where
    TToken: Debug + Clone,
    P1: Parser<TToken, T1, TContext>,
    P2: Parser<TToken, T2, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, (T1, T2)> {
        let (input, r1) = self.0.parse(context.clone(), input)?;
        let (input, r2) = self.1.parse(context, input)?;

        Ok((input, (r1, r2)))
    }
}

impl<TToken, T1, T2, T3, TContext, P1, P2, P3> ParserTuple<TToken, (T1, T2, T3), TContext>
    for (P1, P2, P3)
where
    TToken: Debug + Clone,
    P1: Parser<TToken, T1, TContext>,
    P2: Parser<TToken, T2, TContext>,
    P3: Parser<TToken, T3, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, (T1, T2, T3)> {
        let (input, r1) = self.0.parse(context.clone(), input)?;
        let (input, r2) = self.1.parse(context.clone(), input)?;
        let (input, r3) = self.2.parse(context.clone(), input)?;

        Ok((input, (r1, r2, r3)))
    }
}

impl<TToken, T1, T2, T3, T4, TContext, P1, P2, P3, P4>
    ParserTuple<TToken, (T1, T2, T3, T4), TContext> for (P1, P2, P3, P4)
where
    TToken: Debug + Clone,
    P1: Parser<TToken, T1, TContext>,
    P2: Parser<TToken, T2, TContext>,
    P3: Parser<TToken, T3, TContext>,
    P4: Parser<TToken, T4, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, (T1, T2, T3, T4)> {
        let (input, r1) = self.0.parse(context.clone(), input)?;
        let (input, r2) = self.1.parse(context.clone(), input)?;
        let (input, r3) = self.2.parse(context.clone(), input)?;
        let (input, r4) = self.3.parse(context.clone(), input)?;

        Ok((input, (r1, r2, r3, r4)))
    }
}

impl<TToken, T1, T2, T3, T4, T5, TContext, P1, P2, P3, P4, P5>
    ParserTuple<TToken, (T1, T2, T3, T4, T5), TContext> for (P1, P2, P3, P4, P5)
where
    TToken: Debug + Clone,
    P1: Parser<TToken, T1, TContext>,
    P2: Parser<TToken, T2, TContext>,
    P3: Parser<TToken, T3, TContext>,
    P4: Parser<TToken, T4, TContext>,
    P5: Parser<TToken, T5, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, (T1, T2, T3, T4, T5)> {
        let (input, r1) = self.0.parse(context.clone(), input)?;
        let (input, r2) = self.1.parse(context.clone(), input)?;
        let (input, r3) = self.2.parse(context.clone(), input)?;
        let (input, r4) = self.3.parse(context.clone(), input)?;
        let (input, r5) = self.4.parse(context.clone(), input)?;

        Ok((input, (r1, r2, r3, r4, r5)))
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::combinators::tuple;
    use crate::parser::parser_test_helper;
    use crate::text::char_parser;

    #[test]
    fn tuple_test() {
        parser_test_helper(
            ParserContext::new_with_str("ab", ()),
            &('a', 'b'),
            tuple((char_parser('a'), char_parser('b'))),
        );

        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &('a', 'b', 'c'),
            tuple((char_parser('a'), char_parser('b'), char_parser('c'))),
        );

        parser_test_helper(
            ParserContext::new_with_str("abcd", ()),
            &('a', 'b', 'c', 'd'),
            tuple((
                char_parser('a'),
                char_parser('b'),
                char_parser('c'),
                char_parser('d'),
            )),
        );
    }
}
